"""
Contains functionality for creating spatial zone translation.

Also checks on various things for translations (rounding, slithers).
"""
import logging
from typing import Tuple
import warnings

import geopandas as gpd
import pandas as pd
from caf.space import inputs

##### CONSTANTS #####
LOG = logging.getLogger(__name__)
logging.captureWarnings(True)

##### FUNCTIONS #####


def read_zone_shapefiles(zone_1: inputs.ZoneSystemInfo, zone_2: inputs.ZoneSystemInfo) -> dict:
    """
    Read in zone system shapefiles.

    Reads in shapefiles and sets zone id and area column names, as well as
    matching to same crs. If the provided shapefiles don't contain CRS
    information then they're assumed to be "EPSG:27700".

    Parameters
    ----------
    zone_1: inputs.ZoneSystemInfo
        Info on first zone system
    zone_2: inputs.ZoneSystemInfo
        Info on second zone system

    Returns
    -------
    zones(dict): A nested dictionary containing zones
    for translation. zone_1.name and zone_1.name contain 'Zone'
    (GeoDataFrame) and 'ID_col'(str)
    """
    # create geodataframes from zone shapefiles
    z_1 = gpd.read_file(zone_1.shapefile)
    z_2 = gpd.read_file(zone_2.shapefile)

    z_1 = z_1.dropna(axis=1, how="all")
    z_2 = z_2.dropna(axis=1, how="all")

    LOG.info(
        "Count of %s zones: %s",
        zone_2.name,
        z_2.iloc[:, 0].count(),
    )
    LOG.info(
        "Count of %s zones: %s",
        zone_1.name,
        z_1.iloc[:, 0].count(),
    )

    z_1["area"] = z_1.area
    z_1 = z_1.dropna(subset=["area"])
    z_2["area"] = z_2.area
    z_2 = z_2.dropna(subset=["area"])

    zones = {
        zone_1.name: {
            "Zone": z_1.drop("area", axis=1),
            "ID_col": zone_1.id_col,
        },
        zone_2.name: {
            "Zone": z_2.drop("area", axis=1),
            "ID_col": zone_2.id_col,
        },
    }

    for name, zone in zones.items():
        zone["Zone"].rename(
            columns={zone["ID_col"]: f"{name}_id"},
            inplace=True,
        )
        zone["Zone"][f"{name}_area"] = zone["Zone"].area

        if not zone["Zone"].crs:
            warnings.warn(f"Zone {name} has no CRS, setting crs to EPSG:27700.")
            zone["Zone"].crs = "EPSG:27700"

    return zones


def spatial_zone_correspondence(
    zones: dict,
    zone_1: inputs.ZoneSystemInfo,
    zone_2: inputs.ZoneSystemInfo,
):
    """
    Find the spatial zone correspondence.

    Finds zone correspondence through calculating adjustment factors with areas
    only.

    Parameters
    ----------
    zones: Return value from 'read_zone_shapefiles'.
    zone_1: inputs.ZoneSystemInfo
        Info on first zone system
    zone_2: inputs.ZoneSystemInfo
        Info on second zone system

    Returns
    -------
    GeoDataFrame
    GeoDataFrame with 4 columns: zone 1 IDs, zone 2 IDs, zone 1 to zone
    2 adjustment factor and zone 2 to zone 1 adjustment factor.
    """
    # create geodataframe for intersection of zones
    zone_overlay = gpd.overlay(
        zones[zone_1.name]["Zone"],
        zones[zone_2.name]["Zone"],
        how="intersection",
        keep_geom_type=False,
    ).reset_index()
    zone_overlay.loc[:, "intersection_area"] = zone_overlay.area

    # columns to include in spatial correspondence
    column_list = [
        f"{zone_1.name}_id",
        f"{zone_2.name}_id",
    ]

    # create geodataframe with spatial adjusted factors
    spatial_correspondence = zone_overlay.loc[:, column_list]

    # create geodataframe with spatial adjusted factors
    spatial_correspondence.loc[:, f"{zone_1.name}_to_{zone_2.name}"] = (
        zone_overlay.loc[:, "intersection_area"] / zone_overlay.loc[:, f"{zone_1.name}_area"]
    )
    spatial_correspondence.loc[:, f"{zone_2.name}_to_{zone_1.name}"] = (
        zone_overlay.loc[:, "intersection_area"] / zone_overlay.loc[:, f"{zone_2.name}_area"]
    )

    LOG.info("Unfiltered Spatial Correspondence completed")

    return spatial_correspondence


def find_slithers(
    spatial_correspondence: gpd.GeoDataFrame,
    zone_names: tuple[str, str],
    tolerance: float,
):
    """
    Find overlap areas between zones which are very small slithers.

    Finds slithers and filters them out of the spatial zone correspondence
    GeoDataFrame, and returns the filtered zone correspondence as well as the
    GeoDataFrame wit only the slithers.

    Parameters
    ----------
    spatial_correspondence : GeoDataFrame
        Spatial zone correspondence between zone 1 and zone 2 produced with
        spatial_zone_correspondence
    zone_names: List[str, str]
        List of the zone names that the spatial correspondence was performed
        between
    tolerance : float
        User-defined tolerance for filtering out slithers, must be a float
        between 0 and 1, recommended value is 0.98

    Returns
    -------
    GeoDataFrame, GeoDataFrame
        slithers, GeoDataFrame with all the small zone overlaps, and
        no_slithers, the zone correspondence GeoDataFrame with these zones
        filtered out. slithers isn't usually required but is returned for
        testing.
    """
    LOG.info("Finding Slithers")

    slither_filter = (
        spatial_correspondence[f"{zone_names[0]}_to_{zone_names[1]}"] < (1 - tolerance)
    ) & (spatial_correspondence[f"{zone_names[1]}_to_{zone_names[0]}"] < (1 - tolerance))
    slithers = spatial_correspondence.loc[slither_filter]
    no_slithers = spatial_correspondence.loc[~slither_filter]

    return slithers, no_slithers


def rounding_correction(
    zone_corr: pd.DataFrame, from_zone_name: str, to_zone_name: str
) -> pd.DataFrame:
    """
    Fix error causing negative factors.

    For most translations this function will do almost nothing, but is run
    anyway.

    Parameters
    ----------
    zone_corr (pd.DataFrame): Zone translation dataframe.
    from_zone_name (str): Name of zone_1.
    to_zone_name (str): Name of zone_2.

    Returns
    -------
    pd.DataFrame: The input zone_corr dataframe adjusted to remove errors.
    """

    def calculate_differences(
        frame: pd.DataFrame,
    ) -> Tuple[pd.Series, pd.DataFrame]:
        totals = frame[[from_col, factor_col]].groupby(from_col).sum()
        diffs = (1 - totals).rename(columns={factor_col: "diff"})
        # Convert totals to a Series
        totals = totals.iloc[:, 0]
        return totals, diffs

    from_col = f"{from_zone_name}_id"
    factor_col = f"{from_zone_name}_to_{to_zone_name}"

    counts = zone_corr.groupby(from_col).size()

    # Set factor to 1 for one to one lookups
    zone_corr.loc[zone_corr[from_col].isin(counts[counts == 1].index), factor_col] = 1.0

    # calculate missing adjustments for those that don't have a one to one mapping
    rest_to_round = zone_corr.loc[zone_corr[from_col].isin(counts[counts > 1].index)]
    factor_totals, differences = calculate_differences(zone_corr)

    LOG.info(
        "Adjusting %s correspondence factors for %s which don't sum to exactly 1\n"
        "Difference statistics: max: %.3g, min: %.3g, mean: %.3g, median: %.3g",
        (factor_totals != 1).sum(),
        factor_col,
        differences["diff"].max(),
        differences["diff"].min(),
        differences["diff"].mean(),
        differences["diff"].median(),
    )

    # Calculate factor to adjust the zone correspondence by
    differences.loc[:, "correction"] = 1 + (differences["diff"] / factor_totals)

    # Multiply zone corresondence by the correction factor
    rest_to_round = rest_to_round.merge(
        differences["correction"],
        how="left",
        left_on=from_col,
        right_on=differences.index,
    ).set_index(rest_to_round.index)

    rest_to_round.loc[:, factor_col] = (
        rest_to_round.loc[:, factor_col] * rest_to_round.loc[:, "correction"]
    )

    rest_to_round = rest_to_round.drop(labels="correction", axis=1)

    zone_corr.loc[zone_corr[from_col].isin(rest_to_round[from_col]), :] = rest_to_round

    # Recalculate differences after adjustment
    factor_totals, differences = calculate_differences(zone_corr)

    LOG.info(
        "After adjustment of %s, %s correspondence factors don't sum to exactly 1\n"
        "Difference statistics: max: %.3g, min: %.3g, mean: %.3g, median: %.3g",
        (factor_totals != 1).sum(),
        factor_col,
        differences["diff"].max(),
        differences["diff"].min(),
        differences["diff"].mean(),
        differences["diff"].median(),
    )

    # Check for negative zone correspondences
    negatives = (zone_corr[factor_col] < 0).sum()
    if negatives > 0:
        raise ValueError(f"{negatives} negative correspondence factors for {factor_col}")
    too_big = (zone_corr[factor_col] > 1).sum()
    if too_big > 0:
        raise ValueError(f"{too_big} correspondence factors > 1 for {factor_col}")

    return zone_corr


def round_zone_correspondence(
    zone_corr_no_slithers: pd.DataFrame, zone_names: Tuple[str, str]
):
    """
    Round translation factors.

    Changes zone_1_to_zone_2 adjustment factors such that they sum to 1 for
    every zone in zone 1.

    Parameters
    ----------
    zone_corr_no_slithers : pd.DataFrame
        3 column (zone 1 id, zone 2 id, zone 1 to zone 2) zone correspondence
        DataFrame, with slithers filtered out
    zone_names : List[str, str]
        List of zone 1 and zone 2 names

    Returns
    -------
    pd.DataFrame
        3 column zone correspondence DataFrame with zone_1_to_zone_2 values
        which sum to 1 for each zone 1 id.
    """
    LOG.info(
        "Rounding Zone Correspondences, spatial gaps in the overlap of zone "
        "2 onto zone 1 will be equally distributed between each of zone 2 zones"
    )

    # create rounded zone correspondence
    zone_corr_rounded = rounding_correction(
        zone_corr_no_slithers[
            [
                f"{zone_names[0]}_id",
                f"{zone_names[1]}_id",
                f"{zone_names[0]}_to_{zone_names[1]}",
            ]
        ].copy(),
        *zone_names,
    )

    # Save rounding to final variable to turn to csv
    zone_corr_rounded_both_ways = zone_corr_rounded

    # create rounded zone correspondence for the other direction
    zone_corr_rounded = rounding_correction(
        zone_corr_no_slithers[
            [
                f"{zone_names[0]}_id",
                f"{zone_names[1]}_id",
                f"{zone_names[1]}_to_{zone_names[0]}",
            ]
        ].copy(),
        zone_names[1],
        zone_names[0],
    )

    zone_corr_rounded_both_ways = zone_corr_rounded_both_ways.merge(
        zone_corr_rounded[f"{zone_names[1]}_to_{zone_names[0]}"],
        how="left",
        left_on=zone_corr_rounded_both_ways.index,
        right_on=zone_corr_rounded.index,
    )

    zone_corr_rounded_both_ways = zone_corr_rounded_both_ways.drop(labels="key_0", axis=1)

    return zone_corr_rounded_both_ways


def missing_zones_check(
    zones: dict,
    zone_correspondence: pd.DataFrame,
    zone_1: inputs.ZoneSystemInfo,
    zone_2: inputs.ZoneSystemInfo,
):
    """
    Find missing zones.

    Checks for zone 1 and zone 2 zones missing from zone correspondence.

    Parameters
    ----------
    zones : List[gpd.GeoDataFrame, gpd.GeoDataFrame]
        Zone 1 and zone 2 GeoDataFrames.
    zone_correspondence : pd.DataFrame
        Zone correspondence between zone systems 1 and 2.
    zone_1: inputs.ZoneSystemInfo
        Info on first zone system
    zone_2: inputs.ZoneSystemInfo
        Info on second zone system
    Returns
    -------
    pd.DataFrame
        Zone 1 missing zones.
    pd.DataFrame
        Zone 2 missing zones.
    """
    LOG.info("Checking for missing zones")

    missing_zone_1 = zones[zone_1.name]["Zone"].loc[
        ~zones[zone_1.name]["Zone"][f"{zone_1.name}_id"].isin(
            zone_correspondence[f"{zone_1.name}_id"]
        ),
        f"{zone_1.name}_id",
    ]
    missing_zone_2 = zones[zone_2.name]["Zone"].loc[
        ~zones[zone_2.name]["Zone"][f"{zone_2.name}_id"].isin(
            zone_correspondence[f"{zone_2.name}_id"]
        ),
        f"{zone_2.name}_id",
    ]
    missing_zone_1_zones = pd.DataFrame(
        data=missing_zone_1,
        columns=[f"{zone_1.name}_id"],
    )
    missing_zone_2_zones = pd.DataFrame(
        data=missing_zone_2,
        columns=[f"{zone_2.name}_id"],
    )

    return missing_zone_1_zones, missing_zone_2_zones



"""
    Module for testing the zone_translation module
"""
from pathlib import Path
import pytest
import pandas as pd
import geopandas as gpd

# pylint: disable=import-error, wrong-import-position
from caf.space import zone_translation
from caf.space import inputs


# pylint: enable=import-error,wrong-import-position


@pytest.fixture(name="zone_2_moved", scope="class")
def fixture_zone_2_moved(zone_2_shape, main_dir) -> Path:
    """
    TODO add in testing for this fixture
    Parameters
    ----------
    zone_2_shape
    main_dir

    Returns
    -------

    """
    gdf = gpd.read_file(zone_2_shape)
    gdf.affine_transform(1, 0, 0, 1, 4, 4)
    file = main_dir / "zone_2_altered.shp"
    gdf.to_file(file)
    return file


@pytest.fixture(name="dupe_shapes_config", scope="class")
def fixture_dupe_shapes_config(
    weighted_config: inputs.ZoningTranslationInputs,
) -> inputs.ZoningTranslationInputs:
    """
    A config file for testing translations with one zone the same as
    lower zoning.
    Parameters
    ----------
    weighted_config:
        weighted config to modify and use for a new translation.
    Returns
    -------
        A new config with lower zoning and zone 2 the same.
    """
    config = weighted_config.copy()
    config.zone_2 = weighted_config.lower_zoning
    config.zone_2.name = weighted_config.zone_2.name
    config.lower_zoning.name = weighted_config.zone_2.name
    return config


@pytest.fixture(name="expected_weighted", scope="class")
def fixture_expected_weighted() -> pd.DataFrame:
    """
    The expected output from a weighted translation using vanila inputs.
    Compare to output from weighted translation in testing
    Returns:
        pd.DataFrame: 4 columns of zone_1, zone_2, zone_1_to_zone_2,
        zone_2_to_zone_1
    """
    # fmt: off
    output = pd.DataFrame(
        {
            "zone_1_id": ["A", "A", "A", "A", "B", "B", "C", "C"],
            "zone_2_id": ["Z", "X", "Y", "W", "Z", "X", "Z", "Y"],
            "zone_1_to_zone_2": [
                0.059, 0.176, 0.235, 0.529, 0.263, 0.737, 0.500, 0.500,
            ],
            "zone_2_to_zone_1": [
                0.053, 0.176, 0.235, 1.000, 0.263, 0.824, 0.684, 0.765,
            ],
        }
    )
    # fmt: on
    return output


@pytest.fixture(name="expected_spatial", scope="class")
def fixture_expected_spatial() -> pd.DataFrame:
    """
    The expected output from a weighted translation using vanila inputs.
    Compare to output from weighted translation in testing
    Returns:
        pd.DataFrame: 4 columns of zone_1, zone_2, zone_1_to_zone_2,
        zone_2_to_zone_1
    """
    # fmt: off
    output = pd.DataFrame(
        {
            "zone_1_id": ["A", "A", "A", "A", "B", "B", "C", "C"],
            "zone_2_id": ["Z", "X", "Y", "W", "Z", "X", "Z", "Y"],
            "zone_1_to_zone_2": [
                0.05, 0.2, 0.15, 0.6, 0.2, 0.8, 0.625, 0.375,
            ],
            "zone_2_to_zone_1": [
                0.05, 0.2, 0.25, 1.000, 0.2, 0.8, 0.75, 0.75,
            ],
        }
    )
    # fmt: on
    return output


@pytest.fixture(name="dupe_trans", scope="class")
def fixture_dupe_trans(dupe_shapes_config):
    """
    A weighted zone translation with zone_2 and lower_zone the same.
    Parameters
    ----------
    dupe_shapes_config: see fixture
    Returns
    -------

    """
    trans = zone_translation.ZoneTranslation(dupe_shapes_config).weighted_translation()
    return trans


class TestZoneTranslation:
    """
    Class containing tests for the ZoneTranslation class
    """

    @pytest.mark.parametrize(
        "translation_str", ["spatial_trans", "weighted_trans", "dupe_trans"]
    )
    @pytest.mark.parametrize("origin_zone", [1, 2])
    def test_sum_to_1(self, translation_str: str, origin_zone: int, request):
        """
        Test that translation totals from each zone sum to 1.
        Parameters
        ----------
        translation_str: Used to parametrize and test the different translations
        origin_zone: 1 or 2 for each test zone system/
        request: pytest inbuilt for parametrizing with fixtures

        Returns
        -------

        """
        dic = {1: 2, 2: 1}
        trans = request.getfixturevalue(translation_str)
        summed = trans.groupby(f"zone_{origin_zone}_id").sum()
        rounded = round(summed[f"zone_{origin_zone}_to_zone_{dic[origin_zone]}"], 5).astype(
            "int"
        )
        assert (rounded == 1).all()

    @pytest.mark.parametrize(
        "translation_str", ["spatial_trans", "weighted_trans", "dupe_trans"]
    )
    @pytest.mark.parametrize("col", ["zone_1_to_zone_2", "zone_2_to_zone_1"])
    def test_positive(self, translation_str: str, col: str, request):
        """
        Tests that all translation values are positive
        Parameters
        ----------
        translation_str
        col: column being checked for positives
        request

        Returns
        -------

        """
        trans = request.getfixturevalue(translation_str)
        assert (trans[col] > 0).all()

    @pytest.mark.parametrize("number", [1, 2])
    def test_same_zones(self, spatial_trans, weighted_trans, number: int):
        """
        Test that the two id columns are identical in weighted and
        spatial translations.
        Args:
            spatial_trans (_type_): The spatial translation being checked
            weighted_trans (_type_): Weighted translation being checked
        """
        assert sorted(spatial_trans[f"zone_{number}_id"]) == sorted(
            weighted_trans[f"zone_{number}_id"]
        )

    @pytest.mark.parametrize(
        "expected_str,trans_str",
        [("expected_spatial", "spatial_trans"), ("expected_weighted", "weighted_trans")],
    )
    def test_output(self, trans_str: str, expected_str: str, request):
        """
        Test to see if generated test case zone translations match expected values calculated
        independently.
        Parameters
        ----------
        All provided to request to be read from fixtures.
        trans_str
        expected_str
        request

        Returns
        -------

        """
        trans = request.getfixturevalue(trans_str)
        expected = request.getfixturevalue(expected_str)
        df_1 = trans.groupby(["zone_1_id", "zone_2_id"]).sum().round(3)
        df_1.sort_index(inplace=True)
        df_2 = expected.groupby(["zone_1_id", "zone_2_id"]).sum()
        df_2.sort_index(inplace=True)
        pd.testing.assert_frame_equal(df_1, df_2)



# -*- coding: utf-8 -*-
"""
Module containing ZoneTranslation class.

Class for producing zone translations from a set of inputs, provided by the
ZoneTranslationInputs class in 'inputs'.
"""
import logging
import warnings
import pandas as pd
from caf.space import weighted_funcs, zone_correspondence, inputs

##### CONSTANTS #####
LOG = logging.getLogger(__name__)
logging.captureWarnings(True)


##### CLASSES #####
class ZoneTranslation:
    """
    Store paramaters and create zone translations.

    This is the main class for the caf.space tool. Running it with an
    instance of 'inputs.ZoningTranslationInputs' will return a zone
    translation dataframe.

    Parameters
    ----------
    params: ZoningTranslationInputs
        Params should usually be read in from a yml file using the
        load_yaml method of the ZoningTranslationInputs class. Refer to
        this class for info on parameters.

    Returns
    -------
    Instance of self
        Instance of ZoneTranslation, where a spatial or weighted
        zone translation, dependent on inputs, is stored as
        self.zone_translation
    """

    def __init__(self, params: inputs.ZoningTranslationInputs):
        self.params = params
        self.zone_1 = params.zone_1
        self.zone_2 = params.zone_2
        self.cache_path = params.cache_path
        if params.lower_zoning:
            self.lower_zoning = params.lower_zoning
        if params.method:
            self.method = params.method
        self.tolerance = params.tolerance
        self.rounding = params.rounding
        self.filter_slithers = params.filter_slithers
        self.run_date = params.run_date
        sorted_names = sorted([params.zone_1.name, params.zone_2.name])
        self.names = (sorted_names[0], sorted_names[1])
        self.cacher = self.cache_path / f"{self.names[0]}_{self.names[1]}"
        self.cacher.mkdir(exist_ok=True, parents=True)

    def spatial_translation(self) -> pd.DataFrame:
        """
        Create spatial zone translation.

        Performs zone correspondence between two zoning systems, zone 1 and
        zone 2. Default correspondence is spatial (by zone area), but includes
        options for handling point zones with different data (for example LSOA
        employment data). Also includes option to check adjustment factors from
        zone 1 to zone 2 add to 1.

        Returns
        -------
        Dataframe containing spatial zone translation.
        """
        zones = zone_correspondence.read_zone_shapefiles(self.zone_1, self.zone_2)
        spatial_correspondence = zone_correspondence.spatial_zone_correspondence(
            zones, self.zone_1, self.zone_2
        )
        final_zone_corr = self._slithers_and_rounding(spatial_correspondence)
        # Save correspondence output
        (
            missing_zones_1,
            missing_zones_2,
        ) = zone_correspondence.missing_zones_check(
            zones, final_zone_corr, self.zone_1, self.zone_2
        )

        warnings.warn(f"Missing Zones from 1 : {len(missing_zones_1)}")
        warnings.warn(f"Missing Zones from 2 : {len(missing_zones_2)}")
        log_path = self.cache_path / f"{self.names[0]}_{self.names[1]}"
        log_path.mkdir(exist_ok=True, parents=True)
        log_file = log_path / "missing_zones_log.xlsx"
        with pd.ExcelWriter(
            log_file, engine="openpyxl"
        ) as writer:  # pylint: disable=abstract-class-instantiated
            missing_zones_1.to_excel(
                writer,
                sheet_name=f"{self.names[0]}_missing",
                index=False,
            )
            missing_zones_2.to_excel(
                writer,
                sheet_name=f"{self.names[1]}_missing",
                index=False,
            )
        LOG.info(
            "List of missing zones can be found in log file found here: %s",
            log_file,
        )
        final_zone_corr.to_csv(self.cacher / f"{self.run_date}.csv")
        self.params.save_yaml(self.cacher / f"{self.run_date}.yml")

        return final_zone_corr

    def weighted_translation(self) -> pd.DataFrame:
        """
        Create a weighted zone translation.

        Runs a weighted translation using the zone 1 to lower correspondence
        csv file and the zone 2 to lower correspondence csv file. The type of
        variable to weight the translation by, such as population or
        employment, is chosen by the method variable.

        Parameters
        ----------
        self: for this to run self.params must contain lower zoning and a method.

        Returns
        -------
        weighted_translation: pd.DataFrame
            Weighted Translation between Zone 1 and Zone 2
        """
        LOG.info("Starting weighted translation")
        # Init
        if self.params.method is False:
            raise ValueError("A method must be provided to perform a weighted translation.")
        if self.params.lower_zoning is False:
            raise ValueError("Lower zoning data is required for a weighted translations.")
        cacher_out = self.cacher / self.method
        cacher_out.mkdir(exist_ok=True, parents=False)
        weighted_translation = weighted_funcs.final_weighted(
            self.zone_1, self.zone_2, self.lower_zoning
        )
        weighted_translation = weighted_translation[
            [
                f"{self.names[0]}_to_{self.names[1]}",
                f"{self.names[1]}_to_{self.names[0]}",
            ]
        ]
        weighted_translation.reset_index(inplace=True)

        weighted_translation = self._slithers_and_rounding(weighted_translation)

        column_list = list(weighted_translation.columns)

        summary_table_1 = weighted_translation.groupby(column_list[0])[column_list[2]].sum()
        summary_table_2 = weighted_translation.groupby(column_list[1])[column_list[3]].sum()

        under_1_zones_1 = summary_table_1[summary_table_1 < 0.999999]
        under_1_zones_2 = summary_table_2[summary_table_2 < 0.999999]

        if len(pd.unique(weighted_translation[column_list[0]])) == sum(summary_table_1):
            LOG.info("Split factors add up to 1 for %s", column_list[0])
        else:
            LOG.warning(
                "Split factors DO NOT add up to 1 for %s. CHECK "
                "TRANSLATION IS ACCURATE\n%s",
                column_list[0],
                under_1_zones_1,
            )

        if len(pd.unique(weighted_translation[column_list[1]])) == sum(summary_table_2):
            LOG.info("Split factors add up to 1 for %s", column_list[1])
        else:
            LOG.warning(
                "Split factors DO NOT add up to 1 for %s. CHECK "
                "TRANSLATION IS ACCURATE\n%s",
                column_list[1],
                under_1_zones_2,
            )
        weighted_translation.to_csv(cacher_out / f"{self.params.run_date}.csv", index=False)
        self.params.save_yaml(cacher_out / f"{self.params.run_date}.yml")

        return weighted_translation

    def _slithers_and_rounding(self, translation: pd.DataFrame) -> pd.DataFrame:
        """
        Process slithers and rounding parameters.

        Reads params and filters slithers and rounds outputs if those
        parameters are selected.

        Parameters
        ----------
        translation: The zone translation dataframe for the method to be run on.

        Returns
        -------
        The input dataframe with slithers removed and/or values rounded
        according to input params.
        """
        if self.params.filter_slithers:
            LOG.info("Filtering out small overlaps.")
            (
                _,
                spatial_correspondence_no_slithers,
            ) = zone_correspondence.find_slithers(
                translation, self.names, self.params.tolerance
            )

            if self.params.rounding:
                LOG.info("Checking all adjustment factors add to 1")
                final_zone_corr = zone_correspondence.round_zone_correspondence(
                    spatial_correspondence_no_slithers, self.names
                )
            else:
                final_zone_corr = spatial_correspondence_no_slithers
        else:
            if self.params.rounding:
                LOG.info("Checking all adjustment factors add to 1")
                final_zone_corr = zone_correspondence.round_zone_correspondence(
                    translation, self.names
                )
            else:
                final_zone_corr = translation

        return final_zone_corr



"""Module for metadata. None of this is currently used but hasn't been deleted."""
from __future__ import annotations
import datetime
from pathlib import Path


# Third party imports
# pylint: disable=no-name-in-module
from caf.toolkit import BaseConfig

# pylint: enable=no-name-in-module


class WeightMetadata(BaseConfig):
    """
    Class for storing metadata about weighting date.

    This is not currently used but could be part of a GUI.
    """

    data_col: str
    id_col: str
    year: int
    type: str


class ShapefileMetadata(BaseConfig):
    """
    Store shapefile metadata.

    Class for creating, storing and loading metadata relating to shapefiles
    used for translations. This is not currently used but could be part of a
    GUI.
    """

    name: str
    path: Path
    id_col: str
    weighting_data: WeightMetadata | None = None


class SpatialTransLog(BaseConfig):
    """
    Output log of a spatial translation.

    Mainly used for the lower translations used in weighted translations.

    Parameters
    ----------
    zone_shapefile: Path to the primary zone shapefile used in the
    translation this metadata corresponds to.
    lower_shapefile: Path to the lower shapefile used in the translation
    this metadata corresponds to.
    date: The date this translation took place.
    """

    zone_shapefile: Path
    lower_shapefile: Path
    date: datetime.datetime


class LowerMetadata(BaseConfig):
    """
    Store metadata for all translations between two zones.

    Every time a translation is run between two zones info about that
    translation should be added to an instance of this class saved as
    'metadata.yml'.

    Parameters
    ----------
    translations: A list of SpatialTransLog classes for all translations
    for a given pair of zone systems.
    """

    translations: list[SpatialTransLog]



# -*- coding: utf-8 -*-
"""
Inputs module.

Module containing functionality for storing input parameters and reading config
file. Classes in this module inherit from the BaseConfig class, and are
ultimately used as input parameters for the ZoneTranslation class.
"""

##### IMPORTS #####
from __future__ import annotations

# Standard imports
# pylint: disable=import-error
import logging
import datetime
import fiona
import os
from pathlib import Path
import pandas as pd
from typing import Optional
from pydantic import validator

# Third party imports
from caf.toolkit import BaseConfig

# pylint: enable=import-error
# Local imports

##### CONSTANTS #####
LOG = logging.getLogger(__name__)


class ZoneSystemInfo(BaseConfig):
    """Base class for storing information about a shapefile input.

    Parameters
    ----------
    name : str
        The name of the zone system you are providing. This should
        be as simple as possible, so for and MSOA shapefile, name should
        simply be.
    shapefile : Path
        Path to the shapefile.
    id_col : str
        The name of the unique ID column in your chosen shapefile. This
        can be any column as long as it is unique for each zone in the
        shapefile.
    """

    name: str
    shapefile: Path
    id_col: str

    @validator("id_col")
    def _id_col_in_file(cls, v, values):
        with fiona.collection(values["shapefile"]) as source:
            schema = source.schema
            if v not in schema["properties"].keys():
                raise ValueError(
                    f"The id_col provided, {v}, does not appear"
                    f" in the given shapefile, {'shapefile'}."
                )
        return v

    @validator("shapefile")
    def _path_exists(cls, v):
        """
        Validate a path exists.

        Raises
        ------
        ValueError: Informs user the path given is incorrect.

        Returns
        -------
        Unchanged path if no error is raised.
        """
        if os.path.isfile(v) is False:
            raise ValueError(
                f"The path provided for {v} does not exist."
                "If this path is on a network drive make sure you are connected"
            )
        return v


class LowerZoneSystemInfo(ZoneSystemInfo):
    """
    Lower level zone system input data for `ZoneTranslationInputs`.

    Inherits from ShapefileInfo.

    Parameters
    ----------
    weight_data: Path
        File path to the weighting data for the lower zone system. This
        should be saved as a csv, and only needs two columns (an ID
        column and a column of weighting data)
    data_col: str
        The name of the column in the weighting data csv containing the
        weight data.
    weight_id_col: str
        The name of the columns in the weighting data containing the
        zone ids. This will be used to join the weighting data to the
        lower zoning, so the IDs must match, but the names of the ID
        columns may be different.
    """

    weight_data: Path
    data_col: str
    weight_id_col: str

    def _lower_to_higher(self) -> ZoneSystemInfo:
        return ZoneSystemInfo(name=self.name, shapefile=self.shapefile, id_col=self.id_col)

    @validator("weight_data")
    def _weight_data_exists(cls, v):
        if os.path.isfile(v) is False:
            raise FileNotFoundError(f"The weight data path provided for {v} does not exist.")
        return v

    @validator("data_col", "weight_id_col")
    def _valid_data_col(cls, v, values):
        cols = pd.read_csv(values["weight_data"], nrows=1).columns
        if v not in cols:
            raise ValueError(f"The given col, {v}, does not appear in the weight data.")
        return v


class ZoningTranslationInputs(BaseConfig):
    """
    Class for storing and reading input parameters for `ZoneTranslation`.

    Parameters
    ----------
    zone_1: ZoneSystemInfo
        Zone system 1 information
    zone_2: ZoneSystemInfo
        Zone system 2 information
    lower_zoning: LowerZoneSystemInfo
        Information about the lower zone system, used for performing
        weighted translations. This should be as small a zone system as
        possible relative to zone_1 and zone_2. For spatially weighted
        translations this isn't needed
    cache_path: Path
        File path to a cache of existing translations. This defaults to
        a location on a network drive, and it is best to keep it there,
        but it's more important for weighted translations.
    method: str, optional
        The name of the method used for weighting (e.g. pop or emp).
        This can be anything, but must be included as the tool checks if
        this parameter exists to decide whether to perform a spatial or
        weighted translation.
    tolerance: float, default 0.98
        This is a float less than 1, and defaults to 0.98. If
        filter_slivers (explained below) is chosen, tolerance controls
        how big or small the slithers need to be to be rounded away. For
        most users this can be kept as is.
    rounding: bool, default True
        Select whether or not zone totals will be rounded to 1 after the
        translation is performed. Recommended to keep as True.
    filter_slithers: bool, True
        Select whether very small overlaps between zones will be
        filtered out. This accounts for zone boundaries not aligning
        perfectly when they should between shapefiles, and the tolerance
        for this is controlled by the tolerance parameter. With this
        parameter set to false translations can be a bit messy.
    run_date: str, datetime.datetime.now().strftime("%d_%m_%y")
        When the tool is being run. This is always generated
        automatically and shouldn't be included in the config yaml file.
    """

    # TODO choose and set default path for cache path
    zone_1: ZoneSystemInfo
    zone_2: ZoneSystemInfo
    cache_path: Path
    method: Optional[str] = None
    lower_zoning: Optional[LowerZoneSystemInfo] = None
    tolerance: float = 0.98
    rounding: bool = True
    filter_slithers: bool = True
    run_date: str = datetime.datetime.now().strftime("%d_%m_%y")

    def __post_init__(self) -> None:
        """Make directories if they don't exist."""
        self.cache_path.mkdir(exist_ok=True, parents=True)

    def write_example(self, out_path: Path):
        """
        Write out an example config file.

        When creating a real config, any optional parameters not set should be
        removed entirely, and not just left blank to the right of the colon.

        Parameters
        ----------
        out_path: Path
            The folder the example will be saved in. The file will be
            called 'example.yml'.
        """
        zones = {}
        for i in range(1, 3):
            zones[i] = ZoneSystemInfo(
                name=f"zone_{i}_name",
                shapefile=Path(f"path/to/shapefile_{i}"),
                id_col=f"id_col_for_zone_{i}",
            )
        lower = LowerZoneSystemInfo(
            name="lower_zone_name",
            shapefile=Path("path/to/lower/shapefile"),
            id_col="id_col_for_lower_zone",
            weight_data=Path("path/to/lower/weight/data"),
            data_col="data_col_name",
            weight_id_col="id_col_in_weighting_data",
        )
        ex = ZoningTranslationInputs(
            zone_1=zones[1],
            zone_2=zones[2],
            lower_zoning=lower,
            output_path=r"path\to\output\folder",
            cache_path=r"path\to\cache\folder\defaults\to\ydrive",
            method="OPTIONAL name of method",
        )
        ex.save_yaml(out_path / "example.yml")



"""
Contains functionality for creating weighted translations.

These are called in the 'weighted_trans' method in ZoneTranslation.
"""
##### IMPORTS #####
import logging
import warnings
from functools import reduce
import pandas as pd

# pylint: disable=import-error
import geopandas as gpd

# pylint: enable=import-error

from caf.space import inputs

##### CONSTANTS #####
logging.captureWarnings(True)
LOG = logging.getLogger(__name__)


##### FUNCTIONS #####
def _weighted_lower(
    lower_zoning: inputs.LowerZoneSystemInfo,
) -> gpd.GeoDataFrame:
    """
    Join weighting data to lower zoning shapefile ready to apply.

    Parameters
    ----------
    lower_zoning: Info on lower zoning and weight data

    Returns
    -------
    A lower zoning system with weighting joined to it.
    """
    lower_zone = gpd.read_file(lower_zoning.shapefile)
    lower_zone.set_index(lower_zoning.id_col, inplace=True)
    weighting = pd.read_csv(
        lower_zoning.weight_data,
        index_col=lower_zoning.weight_id_col,
    )
    weighted = lower_zone.join(weighting)
    missing = weighted[lower_zoning.data_col].isna().sum()
    warnings.warn(
        f"{missing} zones do not match up between the lower zoning and weighting data."
    )
    weighted["lower_area"] = weighted.area
    return weighted


def _create_tiles(
    zone_1: inputs.ZoneSystemInfo,
    zone_2: inputs.ZoneSystemInfo,
    lower_zoning: inputs.LowerZoneSystemInfo,
) -> pd.DataFrame:
    """
    Create a spanning set of tiles for the weighted translation.

    Parameters
    ----------
    zone_1: Info on first zone system
    zone_2: Info on second zone system
    lower_zoning: Info on lower zoning and weight data

    Returns
    -------
    A set of weighted tiles used for weighted translation.
    """
    zone_1_gdf = gpd.read_file(zone_1.shapefile)
    zone_2_gdf = gpd.read_file(zone_2.shapefile)

    LOG.info(
        "Count of %s zone: %s",
        zone_2.name,
        zone_2_gdf.iloc[:, 0].count(),
    )
    LOG.info(
        "Count of %s zones: %s",
        zone_1.name,
        zone_1_gdf.iloc[:, 0].count(),
    )

    weighting = _weighted_lower(lower_zoning)
    tiles = reduce(
        lambda x, y: gpd.overlay(x, y, keep_geom_type=True),
        [zone_1_gdf, zone_2_gdf, weighting],
    )
    tiles.overlay_area = tiles.area
    tiles.prop = tiles.overlay_area / tiles.lower_area
    tiles[lower_zoning.data_col] *= tiles.prop
    return tiles[
        [
            zone_1.id_col,
            zone_2.id_col,
            lower_zoning.data_col,
        ]
    ]


def return_totals(frame: pd.DataFrame, id_col: str, data_col: str) -> pd.DataFrame:
    """
    Group df by id_col and sums, keeping data_col.

    Parameters
    ----------
    frame: A dataframe.
    id_col: Column to group by.
    data_col: Column to keep

    Returns
    -------
    Grouped and summed df
    """
    totals = frame.groupby(id_col).sum().loc[:, data_col]
    return totals


def get_weighted_translation(
    zone_1: inputs.ZoneSystemInfo,
    zone_2: inputs.ZoneSystemInfo,
    lower_zoning: inputs.LowerZoneSystemInfo,
) -> pd.DataFrame:
    """
    Create overlap totals for zone systems.

    Creates totals and joins back up.

    Parameters
    ----------
    zone_1: Info on first zone system
    zone_2: Info on second zone system
    lower_zoning: Info on lower zoning and weight data

    Returns
    -------
    Dataframe with columns for overlap total, zone 1 total, zone 2 total
    weights.
    """
    # create a set of spanning weighted, tiles. These tiles will be
    # grouped in different ways to produce the translation.
    tiles = _create_tiles(zone_1, zone_2, lower_zoning)
    # produce total weights by each respective zone system.
    totals_1 = return_totals(tiles, zone_1.id_col, lower_zoning.data_col).to_frame()
    totals_2 = return_totals(tiles, zone_2.id_col, lower_zoning.data_col).to_frame()
    # get values of overlaps between zone systems by grouping by both
    # zone systems and summing.
    overlap = (
        tiles.groupby([zone_1.id_col, zone_2.id_col])
        .sum()
        .loc[:, lower_zoning.data_col]
        .to_frame()
    )
    return overlap.join(totals_1, rsuffix="_1").join(
        totals_2, lsuffix="_overlap", rsuffix="_2"
    )


def final_weighted(
    zone_1: inputs.ZoneSystemInfo,
    zone_2: inputs.ZoneSystemInfo,
    lower_zoning: inputs.LowerZoneSystemInfo,
) -> pd.DataFrame:
    """
    Run functions from module to produce a weighted translation.

    Parameters
    ----------
    zone_1: Info on first zone system
    zone_2: Info on second zone system
    lower_zoning: Info on lower zoning and weight data

    Returns
    -------
    A weighted zone translation DataFrame. This contains more column than the
    final output and will be passed through more checks for slither and
    rounding before being output, according to the input parameters.
    """
    full_df = get_weighted_translation(zone_1, zone_2, lower_zoning)
    full_df[f"{zone_1.name}_to_{zone_2.name}"] = (
        full_df[f"{lower_zoning.data_col}_overlap"] / full_df[f"{lower_zoning.data_col}_1"]
    )
    full_df[f"{zone_2.name}_to_{zone_1.name}"] = (
        full_df[f"{lower_zoning.data_col}_overlap"] / full_df[f"{lower_zoning.data_col}_2"]
    )
    full_df.index.names = [f"{zone_1.name}_id", f"{zone_2.name}_id"]
    return full_df



"""Setup file."""
from setuptools import setup
import versioneer

if __name__ == "__main__":
    setup(
        name="caf.space",
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
    )



"""
Module for testing the weighted_funcs module
"""
from copy import deepcopy
import pytest
import pandas as pd
from caf.space import weighted_funcs


@pytest.fixture(name="weighted", scope="class")
def fixture_weighted(weighted_config):
    """
    Fixture returning a lower zone system with a weighting vector attached to
    it.
    """
    weighted = weighted_funcs._weighted_lower(weighted_config.lower_zoning)
    return weighted


@pytest.fixture(name="tiles", scope="class")
def fixture_tiles(weighted_config):
    """
    Fixture returning tiles from the _create_tiles function
    """
    tiles = weighted_funcs._create_tiles(
        weighted_config.zone_1, weighted_config.zone_2, weighted_config.lower_zoning
    )
    return tiles


@pytest.fixture(name="overlaps", scope="class")
def fixture_overlaps(weighted_config):
    """
    Fixture returning overlaps ond totals.
    """
    overlaps = weighted_funcs.get_weighted_translation(
        weighted_config.zone_1,
        weighted_config.zone_2,
        weighted_config.lower_zoning,
    )
    return overlaps


class TestWeightedLower:
    """
    Class for testing the _weighted_lower function in weighted_funcs
    """

    def test_join(self, weighted):
        """
        Check that weighting is correct after join in the _weighted_lower
        """
        summed = weighted.weight.sum()
        assert summed == 310

    def test_area(self, weighted):
        """
        Check area is correct in weighted lower
        """
        assert (weighted.area == 4).all()

    def test_warning(self, weighted_config, main_dir):
        """
        Check the correct warning is raised from _weighted_lower
        """
        weighting = pd.read_csv(weighted_config.lower_zoning.weight_data)
        weighting.lower_id = range(16)
        weighting_path = main_dir / "mismatched_weighting.csv"
        weighting.to_csv(weighting_path)
        mismatched_config = deepcopy(weighted_config)
        mismatched_config.lower_zoning.weight_data = weighting_path
        with pytest.warns(
            UserWarning,
            match="1 zones do not match up between the lower zoning and weighting data.",
        ):
            weighted_funcs._weighted_lower(mismatched_config.lower_zoning)


class TestCreateTiles:
    """
    Class for testing the _create_tiles function in weighted_funcs
    """

    def test_weight(self, tiles):
        """
        Test the weight of generated tiles.
        """
        summed = tiles.weight.sum()
        assert summed == 310


class TestOverlapsTotals:
    """
    Class for testing the overlaps_and_totals function in weighted_funcs
    """

    def test_sums(self, overlaps):
        """
        Test the weight of overlaps output
        """
        overlap_sum = overlaps.weight_overlap.sum()
        assert overlap_sum == 310


# class Test_Cols_In_Both:
#     def __init__(self, right_in: pd.DataFrame, left_in: pd.DataFrame,
#     cols: list, lower_cols_left: list, lower_cols_right: list):
#         self.left_in = left_in
#         self.right_in = right_in
#         self.cols = cols
#         self.lower_cols_left = lower_cols_left
#         self.lower_cols_right = lower_cols_right
#         self.lis, self.left, self.right,  = weighted_funcs._cols_in_both(left_in, right_in)
#
#     def test_returns_lower(self):
#         assert self.left.columns, self.right.columns == self.lower_cols_left, self.lower_cols_right
#
#     def test_returns_matching(self):
#         assert self.list == self.cols
#
# class Test_Var_Appy:
#     def __init__(self, corr_path: str, weight_path: str, area_corr: pd.DataFrame,
#     missing_zones: int):
#         self.corr_path = corr_path
#         self.weight_path = weight_path
#         self.area_corr = area_corr
#         self.output_corr = weighted_funcs._var_apply(corr_path, weight_path,
#         'var','zone_id','lower_name')
#
#     def test_join(self):
#         assert self.output_corr == self.area_corr
#
#     def test_logging(self):
#         with pytest.warns(UserWarning, match="3 zones are not intersected by target zones"):
#             weighted_funcs._var_apply(self.corr_path, self.weight_path_missing, 'var','zone_id','lower_name'
#             )
#
# class Test_Zone_Split:
#     def __init__(self):
#
